
# Converts a binary file into an intel hex file

# origin:  https://community.silabs.com/s/article/converting-binary-firmware-image-files-to-intel-hex-files-bin-to-hex-x?language=en_US

# input file is always: isettos.bin   output file: os.hex

#Usage: python3 bin2hex_v2.py3



mychar = 0

def ConstructRecord(recordType, address, data):

    record = []

    recordType = recordType & 0xFF
    address = (address >> 8) & 0xFF, address & 0xFF
    numBytes = len(data) & 0xFF

    if (recordType == 0x43):  # this means that mychar occurs a lot
      # will construct this:
      # <:><count><addr><0x43><mychar> 8 * [<population> < 0-8 bytes> ] <checksum>
      newdata = [];
      newdata.append(mychar)
      for i in range(0,8):  # repeats for 0..7
        population = 0  # gets a 1 in every bit position where mychar is in the next 8 bytes
        section = []
        for j in range(0,8):
          # get a byte from data
          byte = data[0]
          data.pop(0)   # remove first item from data
          if (byte != mychar):
            population = population * 2      # shift a 0 into population
            section.append(byte)
          else:
            population = population * 2 + 1  # shift a 1 into population
        # now append population and section to newdata
        newdata.append(population)
        newdata = newdata + section  # join two lists
        #print(population,section) # debug
      # at the end, replace data by newdata
      data = newdata
    
    recordStr = ':'

    if (recordType == 0x79) and (address !=0): # The line ':79' means: 64 zero-bytes (also change in sending program ? seems not needed)
        record.append(0x79) 
       
    else:
        record.append(numBytes)
        record += address
        record.append(recordType)
        record += data
	    
        checksum = 0
        for byte in record:
            checksum += byte
        checksum &= 0xFF
	    
        # Two's complement
        checksum = ~checksum + 1
        checksum &= 0xFF
	    
        record.append(checksum)
	
    for byte in record:
        recordStr += '{:02X}'.format(byte)

    recordStr += '\n'
    
    return recordStr

def convertBinaryToHex(binaryPath, hexPath, ignoreErasedRecords = True):
    
    global mychar

    hexFile = open(hexPath, 'w')
    
    #  hexFile.write(ConstructRecord(0x04, 0x0000, (0x00, 0x00)))  #rjh removed this, 6-1-2025

    address = 0
    with open(binaryPath, 'rb') as binaryFile:
        byte = binaryFile.read(1)
        numBytes = 1
        zero = 0 # new rjh
        firstline=1
        upperaddress =0
        # for short encoding of repeating characters
        type1char=0;
        type2char=1;
        type3char=255;
        type1count=0
        type2count=0
        type3count=0 

        data = []
        while byte != b'':
            
            byte = int.from_bytes(byte, byteorder='big') & 0xFF
            zero = zero | byte  # rjh
            data.append(byte)
            if (byte==type1char):
              type1count +=1
            if (byte==type2char):
              type2count +=1
            if (byte==type3char):
              type3count +=1 
            
            if numBytes >= 64:
                #for val in data:
                #    if val != 0xFF or not ignoreErasedRecords:
                #        hexFile.write(ConstructRecord(0x00, address, data))
                #        break

                # determine if there is an often occurring character
                mychar = 0x55 # indicating nothing
                if (type3count > 10):
                  mychar = type3char
                if (type2count > 10):
                  mychar = type2char
                if (type1count > 10):
                  mychar = type1char

                if (address & 0xf0000 != upperaddress):  #  not in current 64k page
                    upperaddress = address & 0xf0000
                    print('next 64K section')
                    hexFile.write(ConstructRecord(0x04, 0x0000, (0x00, address >> 16)))
                    zero = 0xff  # force full line at next address                    
                
                if (address < 0xC000) | (address >0xFFFF):   # skip the c000-ffff range where the loader lives
                    # new rjh, shorthand for lines with only zeroes, count 77
                    if (zero == 0) and (firstline ==0):
                        hexFile.write(ConstructRecord(0x79, address, data)) #  type 79, 64 zero bytes
                    else:
                        if (mychar == 0x55):
                          hexFile.write(ConstructRecord(0x00, address, data)) # full line of data, normal case
                        else:
                          hexFile.write(ConstructRecord(0x43, address, data)) # full line of data, with mychar occurring often

                    zero = 0
                else:
                    zero = 0xff  # force full line at next address
                data = []
                numBytes = 0
                address += 64  
                type1count=0
                type2count=0
                type3count=0              
                
            numBytes += 1
            
            byte = binaryFile.read(1)
            firstline=0
        
    hexFile.write(ConstructRecord(0x01, 0x0000, []))

    print('end address',address);
        
    hexFile.close()

from optparse import OptionParser

if __name__ == "__main__":
    
    parser = OptionParser(usage="usage: %prog -b <path to binary file> -o <path to hex to be created>")
    parser.add_option("-o", "--hexPath",
                                        action="store",
                                        dest="hexPath",
                                        default="",
                                        type="string",
                                        help="Path to the hex file to be generated")
    parser.add_option("-b", "--binaryPath",
                                        action="store", # optional because action defaults to "store"
                                        dest="binaryPath",
                                        type="string",
                                        default="",
                                        help="Path to the binary file",)
    (options, args) = parser.parse_args()
     
 
    binaryPath = options.binaryPath
    hexPath = options.hexPath

    print("Will convert isettos.bin to os.hex.  version 250513.")

    if not binaryPath:
        #binaryPath = input("Input path to binary file to process:")
        binaryPath = "isettos.bin"
        
    if not hexPath:
        #hexPath = input("Input path to hex file to be generated:")
        hexPath = "os.hex"

    convertBinaryToHex(binaryPath, hexPath)
    
    print("Done! Created hex file: {}".format(hexPath))
    
